# Fifth version of the follow line algorithm
# the lines parameters have been changed to length lk and rotation vector wk

from roblib import *
import csv
import transformations as tf
from parameters import _p_0, _p_0_hat, _v_0, _R_0, _noise_d, _noise_pz, _noise_R, dt, _Marks, \
    _nb_box, _box_k_init, _Pa, _Pb, L_t, _vd, _current, nl, L_ak, L_Rk, _L_dk, _L_wk, _Rk, _lk, β, _noise_vr, _P0,\
    Controller, Riptide, IEKF, Int_Observer, Record, next_line

# Blocs
_cont = Controller()
_riptide = Riptide(_p_0, _v_0, _R_0, _noise_d, _noise_pz, _noise_R)
_kal = IEKF(_P0, _Pa, _Pb, _p_0_hat[0, 0], _p_0_hat[1, 0], _v_0)
_obs = Int_Observer(_noise_d, _noise_pz, _noise_R, _box_k_init, _noise_vr)

# recording lists
L_rec = []

print("simulation start")
for k in range(len(L_t)):

    # measurement and correction
    _y_pz, _y_R, _y_d = _riptide.measure()
    _obs.estim_v(_cont.u[0, 0])
    _obs.estim_p(_y_R, _y_pz, _y_d)

    _kal.update_Ck_IEKF(_obs.p_hat)

    _kal.kalman_correct_IEKF(_y_d, _obs.p_hat)
    _px_hat, _py_hat, _vr_hat = _kal.obs_out()

    _p_hat = np.array([[_px_hat, _py_hat, _y_pz]]).T
    _P_p = np.zeros((3, 3))
    _P_p[0:2, 0:2] = _kal.P[0:2, 0:2]
    _P_p[2, 2] = (_noise_pz.diam() / 2) ** 2

    # control update
    _cont.controller_update(_vd, _Rk, _lk, _p_hat, _y_R, _vr_hat, _obs.flag_circle_behavior)
    _joint = np.concatenate((np.array([[β]]), _cont.u[1:4]), 0)
    # record iteration k
    L_rec.append(
        Record(L_t[k], _riptide.p, _riptide.v, _riptide.R, _kal.px_hat, _kal.py_hat, _obs.p_hat[0, 0], _obs.p_hat[1, 0],
               _y_pz, _cont.u, _P_p,
               _obs.pose_box_list, _y_d, _joint))

    # system update for k+1
    β = β - dt * _cont.u[0, 0]

    _riptide.sys_update(_cont.u, _current)
    _kal.update_Ak(_y_R)
    _kal.kalman_predict(_cont.u, _y_R)

    if next_line(_p_hat, _lk):
        if nl >= len(L_ak[0]) - 2:
            break  # end of the mission
        print("next line")
        nl += 1
        l_old = _lk
        _lk = L_ak[:, nl:nl + 2]
        _Rk = L_Rk[nl]

print("simulation done")
print("saving data")

# save landmark to csv
with open('v7/v7_landmark.csv', 'w', newline='') as csvfile:
    fieldnames = ["lmx", "lmy", "lmz"]
    thewriter = csv.DictWriter(csvfile, fieldnames=fieldnames)
    thewriter.writeheader()

    for lm in _Marks:
        lmx = lm[0]
        lmy = lm[1]
        lmz = lm[2]
        thewriter.writerow({'lmx': lmx,
                            'lmy': lmy,
                            'lmz': lmz, })
# save planning to csv
with open('v7/v7_planning.csv', 'w', newline='') as csvfile:
    fieldnames = ["akx", "aky", "akz", "dk", "wkx", "wky", "wkz"]
    thewriter = csv.DictWriter(csvfile, fieldnames=fieldnames)
    thewriter.writeheader()

    _L_dk.append(0)  # to add the last waypoint
    _L_wk = np.concatenate((_L_wk, np.array([[0, 0, 0]]).T), 1)
    for k in range(len(_L_dk)):
        akx = L_ak[0, k]
        aky = L_ak[1, k]
        akz = L_ak[2, k]
        dk = _L_dk[k]
        wkx = _L_wk[0, k]
        wky = _L_wk[1, k]
        wkz = _L_wk[2, k]
        thewriter.writerow({'akx': akx,
                            'aky': aky,
                            'akz': akz,
                            'dk': dk,
                            'wkx': wkx,
                            'wky': wky,
                            'wkz': wkz})

# save state to csv
with open('v7/v7_state.csv', 'w', newline='') as csvfile:
    fieldnames = ["t", "px", "py", "pz", "vr", "px_hat_kal", "py_hat_kal", "px_hat_int", "py_hat_int", "pz_hat",
                  "r11", "r12", "r13", "r21", "r22", "r23", "r31", "r32", "r33",
                  'q1', 'q2', 'q3', 'q4', 'b', 'u2', 'u3', 'u4',
                  'cov_q1', 'cov_q2', 'cov_q3', 'cov_q4',
                  'cov_l1', 'cov_l2', 'cov_l3']
    for i in range(_nb_box):
        fieldnames += ['x_box_' + str(i), 'y_box_' + str(i), 'z_box_' + str(i), 'lx_box_' + str(i), 'ly_box_' + str(i),
                       'lz_box_' + str(i)]

    thewriter = csv.DictWriter(csvfile, fieldnames=fieldnames)
    thewriter.writeheader()

    for k in range(len(L_rec)):
        t = L_rec[k].t
        px = L_rec[k].p[0, 0]
        py = L_rec[k].p[1, 0]
        pz = L_rec[k].p[2, 0]
        px_hat_kal = L_rec[k].px_hat_kal
        py_hat_kal = L_rec[k].py_hat_kal
        px_hat_int = L_rec[k].px_hat_int
        py_hat_int = L_rec[k].py_hat_int
        pz_hat = L_rec[k].yz
        vr = L_rec[k].vr
        r11 = L_rec[k].R[0, 0]
        r12 = L_rec[k].R[0, 1]
        r13 = L_rec[k].R[0, 2]
        r21 = L_rec[k].R[1, 0]
        r22 = L_rec[k].R[1, 1]
        r23 = L_rec[k].R[1, 2]
        r31 = L_rec[k].R[2, 0]
        r32 = L_rec[k].R[2, 1]
        r33 = L_rec[k].R[2, 2]

        # quaternions
        R_hom = np.concatenate((np.concatenate((L_rec[k].R, np.zeros((3, 1))), 1), np.array([[0., 0., 0., 1.]])))
        quat = tf.quaternion_from_matrix(R_hom)
        q1 = quat[0]
        q2 = quat[1]
        q3 = quat[2]
        q4 = quat[3]

        # control signal
        b = L_rec[k].joint[0, 0]
        u2 = L_rec[k].joint[1, 0]  # top fin
        u3 = L_rec[k].joint[2, 0]  # right fin
        u4 = L_rec[k].joint[3, 0]  # left fin

        # covariance
        P = L_rec[k].P_p
        cov_l, cov_R = np.linalg.eig(P)
        cov_R_hom = np.concatenate((np.concatenate((cov_R, np.zeros((3, 1))), 1), np.array([[0., 0., 0., 1.]])))
        cov_quat = tf.quaternion_from_matrix(cov_R_hom)

        cov_q1 = cov_quat[0]
        cov_q2 = cov_quat[1]
        cov_q3 = cov_quat[2]
        cov_q4 = cov_quat[3]

        cov_l1 = cov_l[0]
        cov_l2 = cov_l[1]
        cov_l3 = cov_l[2]

        row_set = {'t': t, 'px': px, 'py': py, 'pz': pz, 'px_hat_kal': px_hat_kal, 'py_hat_kal': py_hat_kal,
                   'px_hat_int': px_hat_int, 'py_hat_int': py_hat_int, 'pz_hat': pz_hat, 'vr': vr,
                   'r11': r11, 'r12': r12, 'r13': r13, 'r21': r21, 'r22': r22, 'r23': r23, 'r31': r31, 'r32': r32,
                   'r33': r33,
                   'q1': q1, 'q2': q2, 'q3': q3, 'q4': q4, 'b': b, 'u2': u2, 'u3': u3, 'u4': u4,
                   'cov_q1': cov_q1, 'cov_q2': cov_q2, 'cov_q3': cov_q3, 'cov_q4': cov_q4,
                   'cov_l1': cov_l1, 'cov_l2': cov_l2, 'cov_l3': cov_l3}

        # localization
        for i in range(_nb_box):
            _pos_box = L_rec[k].pose_boxes[i]
            X, Y, Z = _pos_box[0], _pos_box[1], _pos_box[2]
            row_set.update({'x_box_' + str(i): X.mid(), 'y_box_' + str(i): Y.mid(), 'z_box_' + str(i): Z.mid(),
                            'lx_box_' + str(i): X.diam(), 'ly_box_' + str(i): Y.diam(), 'lz_box_' + str(i): Z.diam()})

        thewriter.writerow(row_set)
print("done")
